@@ -6,10 +6,12 @@ Markdown cells, output, and metadata are ignored.
 
 import ast
 import json
-import secrets
+import re
+from collections import defaultdict
 from itertools import takewhile
-from typing import TYPE_CHECKING, Dict, Iterator, List, Optional, Tuple
+from typing import TYPE_CHECKING, DefaultDict, Dict, Iterator, List
 
+from nbqa.handle_magics import MagicHandler, MagicSubstitution
 from nbqa.notebook_info import NotebookInfo
 
 if TYPE_CHECKING:
@@ -17,7 +19,8 @@ if TYPE_CHECKING:
 
 CODE_SEPARATOR = "# %%"
 MAGIC = ["%%script", "%%bash"]
-IGNORE_LINE_REPLACEMENT = "pass  # nbqa"
+NEWLINES = defaultdict(lambda: "\n\n")
+NEWLINES["isort"] = "\n"
 
 
 def _is_src_code_indentation_valid(source: str) -> bool:
@@ -47,9 +50,9 @@ def _is_src_code_indentation_valid(source: str) -> bool:
     return valid
 
 
-def _handle_line_magic_indentation(
-    source: List[str], line_magic: str
-) -> Tuple[str, str]:
+def _handle_magic_indentation(
+    source: List[str], line_magic: str, magic_replacement: MagicSubstitution
+) -> str:
     """
     Handle the indentation of line magics. Remove unnecessary indentation.
 
@@ -59,35 +62,36 @@ def _handle_line_magic_indentation(
         Source code of the notebook cell
     line_magic : str
         Line magic present in the notebook cell
-
+    magic_replacement : MagicSubstitution
+        Object containing information on ipython magic replacement.
     Returns
     -------
-    Tuple[str, str]
-        Tuple of replacement line and the original line magic statement.
+    str
+        Replacement line for the original line magic statement.
     """
     leading_space = "".join(takewhile(lambda c: c == " ", line_magic))
-    token = secrets.token_hex(3)
 
     # preserve the leading spaces and check if the syntax is valid
-    line_tokenised = f"{leading_space}{IGNORE_LINE_REPLACEMENT}{token}"
+    replacement = magic_replacement.indent_magic_replacement(leading_space)
 
     if leading_space and not _is_src_code_indentation_valid(
-        "".join(source + [line_tokenised])
+        "".join(source + [replacement])
     ):
         # Remove the line magic indentation assuming these leading spaces
         # lead the source to have invalid indentation
         # Currently we don't check if the original source
         # code itself had IndentationError
-        line_tokenised = line_tokenised.lstrip()
-        line_magic = line_magic.lstrip()
+        replacement = magic_replacement.replacement_line
 
     if line_magic.endswith("\n"):
-        line_tokenised += "\n"
+        replacement += "\n"
 
-    return line_tokenised, line_magic
+    return replacement
 
 
-def _replace_line_magics(source: List[str]) -> Iterator[Tuple[str, Optional[str]]]:
+def _replace_magics(
+    source: List[str], magic_substitutions: List[MagicSubstitution]
+) -> Iterator[str]:
     """
     Replace IPython line magics with valid python code.
 
@@ -95,22 +99,30 @@ def _replace_line_magics(source: List[str]) -> Iterator[Tuple[str, Optional[str]
     ----------
     source
         Source from notebook cell.
+    magic_substitutions
+        List to store all the ipython magics substitutions
 
     Yields
     ------
     str
-        Lines from cell, with line magics replaced with python code
+        Line from cell, with line magics replaced with python code
     """
     for line_no, line in enumerate(source):
-        if line.lstrip().startswith(("!", "%", "?")) or line.rstrip().endswith("?"):
-            yield _handle_line_magic_indentation(source[:line_no], line)
-        else:
-            yield line, None
+        trimmed_line: str = line.strip()
+        if MagicHandler.is_ipython_magic(trimmed_line):
+            magic_handler = MagicHandler.get_magic_handler(trimmed_line)
+            magic_substitution = magic_handler.replace_magic(trimmed_line)
+            magic_substitutions.append(magic_substitution)
+            line = _handle_magic_indentation(source[:line_no], line, magic_substitution)
+
+        yield line
 
 
 def _parse_cell(
     source: List[str],
-    temporary_lines: Dict[str, str],
+    cell_number: int,
+    temporary_lines: Dict[int, List[MagicSubstitution]],
+    command: str,
 ) -> str:
     """
     Parse cell, replacing line magics with python code as placeholder.
@@ -119,19 +131,27 @@ def _parse_cell(
     ----------
     source
         Source from notebook cell.
+    cell_number
+        Number identifying the notebook cell.
     temporary_lines
-        Mapping from placeholder python code to the original statement(line magic).
+        Mapping to store the cell number to all the ipython magics replaced in those cells.
+    command
+        The third-party tool being run.
 
     Returns
     -------
     str
         Parsed cell.
     """
-    parsed_cell = f"\n{CODE_SEPARATOR}\n"
-    for new, old in _replace_line_magics(source):
-        parsed_cell += new
-        if old is not None:
-            temporary_lines[new] = old
+    substituted_magics: List[MagicSubstitution] = []
+    parsed_cell = f"{NEWLINES[command]}{CODE_SEPARATOR}\n"
+
+    for parsed_line in _replace_magics(source, substituted_magics):
+        parsed_cell += parsed_line
+
+    if substituted_magics:
+        temporary_lines[cell_number] = substituted_magics
+
     parsed_cell = parsed_cell.rstrip("\n") + "\n"
     return parsed_cell
 
@@ -157,9 +177,7 @@ def _should_ignore_code_cell(source: List[str], ignore_cells: List[str]) -> bool
 
 
 def main(
-    notebook: "Path",
-    temp_python_file: "Path",
-    ignore_cells: List[str],
+    notebook: "Path", temp_python_file: "Path", ignore_cells: List[str], command: str
 ) -> NotebookInfo:
     """
     Extract code cells from notebook and save them in temporary Python file.
@@ -172,6 +190,8 @@ def main(
         Temporary Python file to save converted notebook in.
     ignore_cells
         Extra cells which nbqa should ignore.
+    command
+        The third-party tool being run.
 
     Returns
     -------
@@ -186,7 +206,7 @@ def main(
     line_number = 0
     cell_number = 0
     trailing_semicolons = set()
-    temporary_lines: Dict[str, str] = {}
+    temporary_lines: DefaultDict[int, List[MagicSubstitution]] = defaultdict(list)
     code_cells_to_ignore = set()
 
     for cell in cells:
@@ -197,21 +217,22 @@ def main(
                 code_cells_to_ignore.add(cell_number)
                 continue
 
-            parsed_cell = _parse_cell(cell["source"], temporary_lines)
-            result.append(parsed_cell)
-            split_parsed_cell = parsed_cell.splitlines()
+            parsed_cell = _parse_cell(
+                cell["source"], cell_number, temporary_lines, command
+            )
             cell_mapping.update(
                 {
                     j + line_number + 1: f"cell_{cell_number}:{j}"
-                    for j in range(len(split_parsed_cell))
+                    for j in range(len(parsed_cell.splitlines()))
                 }
             )
             if parsed_cell.rstrip().endswith(";"):
                 trailing_semicolons.add(cell_number)
-            line_number += len(split_parsed_cell)
+            result.append(re.sub(r";(\s*)$", "\\1", parsed_cell))
+            line_number += len(parsed_cell.splitlines())
 
     with open(str(temp_python_file), "w") as handle:
-        handle.write("".join(result)[len("\n") :])
+        handle.write("".join(result).lstrip("\n"))
 
     return NotebookInfo(
         cell_mapping, trailing_semicolons, temporary_lines, code_cells_to_ignore
